from pymoo.optimize import minimize
from pymoo.model.problem import Problem
import numpy as np
from pymoo.factory import get_algorithm, get_termination, get_reference_directions

REF_DIRS = get_reference_directions("das-dennis", 1, n_partitions=6)


def pymoo_brkga_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="brkga", n_dim=n_dim, with_count=with_count)


def pymoo_nelder_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nelder-mead", n_dim=n_dim, with_count=with_count)


def pymoo_cmaes_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="cmaes", n_dim=n_dim, with_count=with_count)


def pymoo_nsga2_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga2", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pymoo_rnsga2_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="rnsga2", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pymoo_rnsga3_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="rnsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pymoo_unsga3_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="unsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pymoo_moead_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="moead", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pymoo_pattern_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="pattern-search", n_dim=n_dim, with_count=with_count)


def pymoo_ctaea_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="ctaea", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pymoo_nsga3_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pymoo_de_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga3", n_dim=n_dim, with_count=with_count)


def pymoo_cube(objective, n_trials, method_name, n_dim, with_count, ref_dirs=None):

    class ObjectiveProblem(Problem):

        def __init__(self):
            super().__init__(n_var=n_dim, n_obj=1, n_constr=0, xl=0.0, xu=1.0)
            self.feval_count = 0

        def _evaluate(self, x, out, *args, **kwargs):
            """ vectorized  """
            self.feval_count = self.feval_count + len(x)
            out["F"] = np.array([objective(u) for u in x])

    try:
        algorithm = get_algorithm(method_name, ref_dirs=ref_dirs)
    except ValueError:
        algorithm = get_algorithm(method_name)
    termination = get_termination("n_eval", n_trials)
    problem = ObjectiveProblem()

    result = minimize(problem=problem,
                      algorithm=algorithm,
                      termination=termination,
                      seed=None,
                      verbose=False,
                      display=None,
                      callback=None,
                      return_least_infeasible=False,
                      save_history=False
                      )
    best_val = result.F[0]
    best_x = result.X.tolist()
    return (best_val, best_x, problem.feval_count) if with_count else (best_val, best_x)

PYMOO_CANDIDATES = [ pymoo_de_cube, pymoo_nsga2_cube, pymoo_rnsga2_cube,
                     pymoo_nelder_cube, pymoo_ctaea_cube,
                     pymoo_nsga3_cube, pymoo_rnsga3_cube, pymoo_unsga3_cube,
                     pymoo_pattern_cube, pymoo_brkga_cube, pymoo_nsga2_cube ]

PYMOO_OPTMIZERS = [ pymoo_nsga2_cube, pymoo_nelder_cube, pymoo_nsga3_cube,
                    pymoo_unsga3_cube,pymoo_pattern_cube, pymoo_brkga_cube,
                    pymoo_nsga2_cube]

# TODO: See why ['pymoo_ctaea_cube', 'pymoo_de_cube', 'pymoo_rnsga2_cube', 'pymoo_rnsga3_cube'] are broken sometimes

if __name__ == '__main__':
    from timemachines.optimizers.objectives import OBJECTIVES
    always_working = PYMOO_CANDIDATES
    broken = set()
    for objective in OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in PYMOO_CANDIDATES:
            try:
                print(optimizer(objective, n_trials=50, n_dim=5, with_count=True))
            except Exception as e:
                print(e)
                broken.add(optimizer)
                always_working.remove(optimizer)
    print(' ')
    print('Sometimes broken: ')
    print([b.__name__ for b in broken])
    print(' ')
    print('Alway working: ')
    print([b.__name__ for b in always_working])




from hyperopt import fmin, hp, tpe, Trials
from hyperopt.tpe import suggest as tpe_suggest
from hyperopt.rand import suggest as rand_suggest
from hyperopt.atpe import suggest as atpe_suggest


def hyperopt_cube(objective, n_trials, n_dim, with_count=False, algo=None):
    """ Minimize a function on the cube using HyperOpt, and audit # of function calls
       :param objective:    function on (0,1)^n_dim
       :param n_trials:     Guideline for function evaluations
       :param n_dim:
       :param with_count:
       :return:
    """
    assert algo is not None, 'provide algo'
    hp_space = dict([('u' + str(i), hp.uniform('u' + str(i), 0, 1)) for i in range(n_dim)])

    global feval_count
    feval_count = 0

    def _objective(hps):
        global feval_count
        feval_count += 1
        us = [hps['u' + str(i)] for i in range(n_dim)]
        return objective(us)

    trls = Trials()
    res = fmin(_objective, space=hp_space, algo=tpe.suggest, trials=trls, max_evals=n_trials, show_progressbar=False)
    best_x = [trls.best_trial['misc']['vals']['u' + str(i)][0] for i in range(n_dim)]
    best_val = trls.best_trial['result']['loss']
    return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


def hyperopt_atpe_cube(objective, n_trials, n_dim, with_count=False):
    return hyperopt_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, algo=atpe_suggest)


def hyperopt_tpe_cube(objective, n_trials, n_dim, with_count=False):
    return hyperopt_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, algo=tpe_suggest)


def hyperopt_rand_cube(objective, n_trials, n_dim, with_count=False):
    return hyperopt_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, algo=rand_suggest)


HYPEROPT_OPTIMIZERS = [hyperopt_atpe_cube, hyperopt_rand_cube, hyperopt_tpe_cube]

if __name__ == '__main__':
    from timemachines.optimizers.objectives import OBJECTIVES

    for objective in OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in HYPEROPT_OPTIMIZERS:
            print(optimizer(objective, n_trials=50, n_dim=5, with_count=True))



from timemachines.conventions import to_space, Y_TYPE, from_space, dimension
from timemachines.evaluation import evaluate_energy, evaluate_mean_squared_error
from timemachines.optimizers.alloptimizers import OPTIMIZERS
from timemachines.skaters.pmd import pmd_auto
import traceback

def optimize(f, ys:[Y_TYPE],
             evaluator,
             optimizer,
             n_trials,
             n_dim,        # Which dimension to search in
             with_count=False,
             **kwargs)->(float, float):
    """ Returns best r """

    def objective(u:[float]):
        r = from_space(u)
        return evaluator(f=f,ys=ys,r=r,**kwargs)

    a_test = objective(u=[0.5]*n_dim)  # Fail fast with easier trace TODO: remove
    return optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count)



if __name__=='__main__':
    from timemachines.synthetic import brownian_with_exogenous
    from timemachines.evaluation import EVALUATORS
    broken = list()
    for optimizer in OPTIMIZERS:
        print(' ')
        print(optimizer.__name__)
        for evaluator in EVALUATORS:
            try:
                print( evaluator.__name__, optimize(f=pmd_auto,ys=brownian_with_exogenous(n=120),
                      n_trials=5, n_dim=3, n_burn=20, optimizer=optimizer,
                      evaluator=evaluator, with_count=True))
            except Exception as e:
                traceback.print_tb()
                broken.append( (optimizer.__name__,evaluator.__name__))
    print(' ')
    print('Broken : ')
    print(broken)




from scipy.optimize import minimize
from timemachines.optimizers.objectives import OBJECTIVES
global feval_count
feval_count=0

# TODO: standardize with SHGO and remove duplicate code
MINIMIZER_KWARGS = {'slqsp': {'method': 'SLQSP',
                              'max_iter': 10},
                    'powell': {'method': 'Powell',
                               'max_iter': 10},
                    'nelder': {'method': 'Nelder-Mead',
                               'maxiter': 10},
                    'dogleg': {'method': 'dogleg',
                               'maxiter': 10}
                    }


def scipy_cube(objective, n_trials, n_dim, with_count=False, method=None):
    bounds = [(0,1) ]*n_dim

    options = MINIMIZER_KWARGS[method]

    global feval_count
    feval_count = 0

    def _objective(x):
        global feval_count
        feval_count +=1
        return objective(list(x))

    result = minimize(_objective, x0=[0]*n_dim, method='powell',bounds=bounds, options={'maxfev':n_trials,'maxiter':n_trials})
    best_x = result.x.tolist()
    best_val = _objective(result.x)
    return (best_val, best_x,  feval_count) if with_count else (best_val, best_x)


def scipy_slqsp_cube(objective, n_trials, n_dim, with_count=False ):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='slqsp')


def scipy_powell_cube(objective, n_trials, n_dim, with_count=False):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='powell')


def scipy_nelder_cube(objective, n_trials, n_dim, with_count=False):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='nelder')


def scipy_dogleg_cube(objective, n_trials, n_dim, with_count=False):
    return scipy_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='dogleg')


SCIPY_OPTIMIZERS = [ scipy_slqsp_cube, scipy_powell_cube, scipy_nelder_cube, scipy_dogleg_cube ]


if __name__ == '__main__':
    for objective in OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in SCIPY_OPTIMIZERS:
            print((optimizer.__name__,(optimizer(objective, n_trials=200, n_dim=6, with_count=True))))

